@@ -14,6 +14,7 @@
 
 #include <json.h>
 
+#include "cleanup.h"
 #include "fabrics.h"
 #include "log.h"
 #include "private.h"
@@ -127,7 +128,7 @@ static void json_parse_port(nvme_subsystem_t s, struct json_object *port_obj)
 
 static void json_parse_subsys(nvme_host_t h, struct json_object *subsys_obj)
 {
-	struct json_object *nqn_obj, *port_array;
+	struct json_object *nqn_obj, *app_obj, *port_array;
 	nvme_subsystem_t s;
 	const char *nqn;
 	int p;
@@ -137,6 +138,12 @@ static void json_parse_subsys(nvme_host_t h, struct json_object *subsys_obj)
 		return;
 	nqn = json_object_get_string(nqn_obj);
 	s = nvme_lookup_subsystem(h, NULL, nqn);
+	if (!s)
+		return;
+	app_obj = json_object_object_get(subsys_obj, "application");
+	if (app_obj)
+		nvme_subsystem_set_application(s, json_object_get_string(app_obj));
+
 	port_array = json_object_object_get(subsys_obj, "ports");
 	if (!port_array)
 		return;
@@ -183,39 +190,42 @@ static void json_parse_host(nvme_root_t r, struct json_object *host_obj)
 	}
 }
 
+static DEFINE_CLEANUP_FUNC(cleanup_tokener, json_tokener *, json_tokener_free)
+#define _cleanup_tokener_ __cleanup__(cleanup_tokener)
+
 static struct json_object *parse_json(nvme_root_t r, int fd)
 {
 	char buf[JSON_FILE_BUF_SIZE];
-	struct json_object *obj = NULL;
-	struct printbuf *pb;
-	json_tokener *tok = NULL;
+	struct json_object *obj;
+	char *str = NULL;
+	_cleanup_tokener_ json_tokener *tok = NULL;
 	int ret;
+	_cleanup_free_ void *ptr = NULL;
+	int len = 0;
 
-	pb = printbuf_new();
-	if (!pb)
+	while ((ret = read(fd, buf, JSON_FILE_BUF_SIZE)) > 0) {
+		str = realloc(ptr, len + ret);
+		if (!str)
+			return NULL;
+		memcpy(&str[len], buf, ret);
+		len += ret;
+		ptr = str;
+	}
+
+	if (ret < 0 || !len)
 		return NULL;
 
-	while ((ret = read(fd, buf, JSON_FILE_BUF_SIZE)) > 0)
-		printbuf_memappend(pb, buf, ret);
-
-	if (ret < 0)
-		goto out;
-
 	tok = json_tokener_new_ex(JSON_TOKENER_DEFAULT_DEPTH);
 	if (!tok)
-		goto out;
+		return NULL;
 
 	/* Enforce correctly formatted JSON */
 	tok->flags = JSON_TOKENER_STRICT;
 
-	obj = json_tokener_parse_ex(tok, pb->buf, printbuf_length(pb));
+	obj = json_tokener_parse_ex(tok, str, len);
 	if (!obj)
 		nvme_msg(r, LOG_DEBUG, "JSON parsing failed: %s\n",
 			 json_util_get_last_err());
-out:
-	if (tok)
-		json_tokener_free(tok);
-	printbuf_free(pb);
 
 	return obj;
 }
@@ -325,21 +335,21 @@ static void json_update_port(struct json_object *ctrl_array, nvme_ctrl_t c)
 	 * Store the keyring description in the JSON config file.
 	 */
 	if (cfg->keyring) {
-		char *desc = nvme_describe_key_serial(cfg->keyring);
+		_cleanup_free_ char *desc =
+			nvme_describe_key_serial(cfg->keyring);
 
 		if (desc) {
 			json_object_object_add(port_obj, "keyring",
 					       json_object_new_string(desc));
-			free(desc);
 		}
 	}
 	if (cfg->tls_key) {
-		char *desc = nvme_describe_key_serial(cfg->tls_key);
+		_cleanup_free_ char *desc =
+			nvme_describe_key_serial(cfg->tls_key);
 
 		if (desc) {
 			json_object_object_add(port_obj, "tls_key",
 					       json_object_new_string(desc));
-			free(desc);
 		}
 	}
 
@@ -350,7 +360,7 @@ static void json_update_subsys(struct json_object *subsys_array,
 			       nvme_subsystem_t s)
 {
 	nvme_ctrl_t c;
-	const char *subsysnqn = nvme_subsystem_get_nqn(s);
+	const char *subsysnqn = nvme_subsystem_get_nqn(s), *app;
 	struct json_object *subsys_obj = json_object_new_object();
 	struct json_object *port_array;
 
@@ -360,6 +370,10 @@ static void json_update_subsys(struct json_object *subsys_array,
 
 	json_object_object_add(subsys_obj, "nqn",
 			       json_object_new_string(subsysnqn));
+	app = nvme_subsystem_get_application(s);
+	if (app)
+		json_object_object_add(subsys_obj, "application",
+				       json_object_new_string(app));
 	port_array = json_object_new_array();
 	nvme_subsystem_for_each_ctrl(s, c) {
 		json_update_port(port_array, c);
